// © 2022, CGVR (https://cgvr.informatik.uni-bremen.de/),
// Author: Andre Mühlenbrock (muehlenb@uni-bremen.de)

// Include the GUI:
#define IMGUI_DEFINE_MATH_OPERATORS
#include <imgui_internal.h>
#include <imgui.h>
#include <backends/imgui_impl_glfw.h>
#include <backends/imgui_impl_opengl3.h>

// Include functions to display Mat4f and Vec4f in the GUI:
# include <imgui_cg1_helpers.h>

// Include OpenGL3.3 Core functions:
#include <glad/glad.h>

// Include GLFW/OpenGL:
#include <GLFW/glfw3.h>

// Include min, max functions:
#include <algorithm>

// Include IO utilities:
#include <iostream>

// Include Vec4f class:
#include "src/math/Vec4f.h"

// Include Mat4f class:
#include "src/math/Mat4f.h"

// Include our shader wrapper class:
#include "src/Shader.h"

// Include our mesh class:
#include "src/Mesh.h"

// Include our texture fbo class:
#include "src/TextureFBO.h"

// Include our new SpinningTop class:
#include "src/game/StaticMeshActor.h"

/**
 * Initializes the application including the GUI
 */
int main(int, char**)
{
    // Setup window:
    glfwSetErrorCallback([](int error, const char* description) {
        fprintf(stderr, "Glfw Error %d: %s\n", error, description);
    });

    if (!glfwInit())
        return 1;

    // Decide GL+GLSL versions:
    const char* glsl_version = "#version 330 core";
    glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
    glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
    glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);
    glfwWindowHint(GLFW_OPENGL_FORWARD_COMPAT, GL_TRUE);
    glfwWindowHint(GLFW_SAMPLES, 4);
    glfwWindowHint(GLFW_MAXIMIZED, true);

    // Create window with graphics context:
    GLFWwindow* window = glfwCreateWindow(860, 560, "Computergraphik 1", NULL, NULL);
    if (window == NULL)
        return 1;

    glfwMakeContextCurrent(window);
    glfwSwapInterval(1);

    // Initialize GLAD functions:
    if (!gladLoadGLLoader((GLADloadproc)glfwGetProcAddress))
    {
        std::cout << "Failed to initialize GLAD" << std::endl;
        return -1;
    }

    // Setup Dear ImGui context:
    IMGUI_CHECKVERSION();
    ImGui::CreateContext();
    ImGuiIO& io = ImGui::GetIO(); (void)io;
    io.ConfigFlags |= ImGuiConfigFlags_NavEnableKeyboard;     // Enable Keyboard Controls

    // Setup Dear ImGui style:
    ImGui::StyleColorsDark();

    // Setup Platform/Renderer backends:
    ImGui_ImplGlfw_InitForOpenGL(window, true);
    ImGui_ImplOpenGL3_Init(glsl_version);

    // Defines the with of the left menu:
    int menuWidth = 260;

    // Load Roboto Font:
    io.Fonts->AddFontFromFileTTF(CMAKE_SOURCE_DIR "/lib/imgui-1.88/misc/fonts/Roboto-Medium.ttf", 16.0f);

    Shader immediateShader = Shader(CMAKE_SOURCE_DIR "/shader/immediateLighting.vert", CMAKE_SOURCE_DIR "/shader/immediateLighting.frag");
    Shader deferredPass1Shader = Shader(CMAKE_SOURCE_DIR "/shader/deferredLightingPass1.vert", CMAKE_SOURCE_DIR "/shader/deferredLightingPass1.frag");
    Shader deferredPass2Shader = Shader(CMAKE_SOURCE_DIR "/shader/deferredLightingPass2.vert", CMAKE_SOURCE_DIR "/shader/deferredLightingPass2.frag");

    // Enable depth test:
    glEnable(GL_DEPTH_TEST);

    // Last mouse position:
    ImVec2 lastMousePosition = ImGui::GetMousePos();

    // We define a world:
    World world;

    // These variables sum the mouse movements in x and y direction
    // while the mouse was pressed in window coordinates (i.e. pixels):
    float viewPositionX = 0.f;
    float viewPositionY = -0.5f;
    float viewDistance = 5.f;

    // Numbers of lights in the scene:
    int lightNum = 1;

    // Show debug lights:
    bool showLights = true;

    // Animate lights:
    bool animateLights = true;

    // DeferredLighting active?
    bool useDeferredLighting = false;

    // Rendering time
    float worldRenderingTime = 0.f;

    // Create our FBO for the first pass including two textures:
    TextureFBO deferredLightingFBO(std::vector<TextureType>({
        TextureType(GL_RGBA16F,  GL_RGBA, GL_FLOAT),                                // Position texture
        TextureType(GL_RGBA16F,  GL_RGBA, GL_FLOAT),                                // Normal texture
        TextureType(GL_RGBA, GL_RGBA, GL_UNSIGNED_BYTE),                            // Color texture
        TextureType(GL_DEPTH_COMPONENT24, GL_DEPTH_COMPONENT, GL_UNSIGNED_BYTE),    // Depth texture type
    }));

    // Creates a simple quad mesh for rendering from [-1, 1].
    std::vector<Triangle> triangles;
    triangles.push_back(Triangle(Vertex(Vec4f(-1.0f, -1.0f, -1.0f)), Vertex(Vec4f(1.0f, 1.0f, -1.0f)), Vertex(Vec4f(1.0f, -1.0f, -1.0f))));
    triangles.push_back(Triangle(Vertex(Vec4f(-1.0f, -1.0f, -1.0f)), Vertex(Vec4f(-1.0f, 1.0f, -1.0f)), Vertex(Vec4f(1.0f, 1.0f, -1.0f))));
    Mesh quadMesh(triangles);

    // Main loop which is executed every frame until the window is closed:
    while (!glfwWindowShouldClose(window))
    {
        // Processes all glfw events:
        glfwPollEvents();

        // Get the size of the window:
        int display_w, display_h;
        glfwGetFramebufferSize(window, &display_w, &display_h);

        // Subtract right menu:
        display_w -= menuWidth;

        // Create the GUI:
        {
            // Start the Dear ImGui frame
            ImGui_ImplOpenGL3_NewFrame();
            ImGui_ImplGlfw_NewFrame();
            ImGui::NewFrame();

            // Create Window with settings:
            ImGui::SetNextWindowPos(ImVec2(0,0));
            ImGui::SetNextWindowSize(ImVec2(float(menuWidth),float(display_h)));
            ImGui::Begin("Einstellungen", nullptr, ImGuiWindowFlags_NoMove | ImGuiWindowFlags_NoCollapse | ImGuiWindowFlags_NoResize);

            ImGui::DragInt("Lichter", &lightNum, 0.5f, 1, 100, "%d");
            ImGui::Separator();
            ImGui::Checkbox("Lichter anzeigen", &showLights);
            ImGui::Checkbox("Lichter animieren", &animateLights);
            ImGui::Separator();
            ImGui::Checkbox("Deferred Lighting aktivieren", &useDeferredLighting);
            ImGui::Separator();
            ImGui::Text(" ");
            ImGui::Separator();
            ImGui::Text("Time: %.3f ms", worldRenderingTime);
            ImGui::Separator();

            ImGui::End();
        }

        // Ensure that the number of lights is between 1 and 100:
        lightNum = std::max(std::min(lightNum, 100),1);

        // If mouse is down, simply add the difference to viewPositionX and viewPositionY:
        if(ImGui::IsMouseDown(ImGuiMouseButton_Left) && !ImGui::GetIO().WantCaptureMouse){
            // Calculate delta between current mouse postion and the mouse position in the
            // last frame:
            ImVec2 mouseDelta = ImGui::GetMousePos() - lastMousePosition;

            // Add the delta to mouseMovementX and mouseMovementY:
            viewPositionX -= mouseDelta.x * 0.002f;
            viewPositionY -= mouseDelta.y * 0.002f;
        }

        // If mouse wheel moved, change the view distance to the scene mid point:
        if(ImGui::GetIO().MouseWheel != 0){
            viewDistance *= ImGui::GetIO().MouseWheel > 0 ? 0.9f : 1.1f;
        }

        // Generate a rotation matrix (which is changed when the mouse is dragged):
        Mat4f rotationMat({cos(viewPositionX), -sin(viewPositionY) * -sin(viewPositionX), cos(viewPositionY) * -sin(viewPositionX), 0, 0, cos(viewPositionY), sin(viewPositionY), 0, sin(viewPositionX), -sin(viewPositionY) * cos(viewPositionX), cos(viewPositionY) * cos(viewPositionX), 0, 0,0,0,1});

        // Store this mouse position for the next frame:
        lastMousePosition = ImGui::GetMousePos();

        // Setup the GL viewport
        glViewport(menuWidth, 0, display_w, display_h);
        glClearColor(0.f, 0.f, 0.f, 1.0f);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        // Calculate aspect ratio (might be used for perspective transformation):
        float aspectRatio = float(display_w) / display_h;

        // Ensure that both sides of the triangle are rendered:
        glEnable(GL_CULL_FACE);
        glCullFace(GL_FRONT);

        // Get projection matrix:
        Mat4f projection = Mat4f::perspectiveTransformation(aspectRatio);

        // Generate the view matrix by using the rotation matrix and translate the camera
        // by the distance to the scene mid point:
        Mat4f view = Mat4f::translation(0,0,viewDistance) * rotationMat * Mat4f::translation(0,-1,0);

        // If lights should be animated, call the simulate-function
        // of the world where the light is animated:
        if(animateLights){
            world.simulate();
        }

        // Start measuring time:
        GLuint timingQuery;
        glGenQueries(1, &timingQuery);
        glBeginQuery(GL_TIME_ELAPSED, timingQuery);

        // Select lighting path:
        if(!useDeferredLighting){
            // Use forward shader:
            immediateShader.bind();

            // Bind light information to the forward / immediate shader:
            world.bindLights(view, immediateShader, lightNum);

            // Render the world and simulate and render all objecst:
            world.render(view, projection, immediateShader);
        } else {
            // Render in a first pass to the deferredLightingFBO:
            {
                // Bind the deferredLightingFBO so we render to textures first:
                deferredLightingFBO.bind(display_w, display_h);

                // Bind the defferedPass1Shader:
                deferredPass1Shader.bind();

                // Setup viewport and clear color & depth image:
                glViewport(0, 0, display_w, display_h);
                glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
                glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

                // Render scene without binding lights beforehand:
                world.render(view, projection, deferredPass1Shader);
            }

            // Render in a second pass to the screen:
            {
                // Activate the default FBO to render directly to the screen:
                glBindFramebuffer(GL_FRAMEBUFFER, 0);
                glViewport(menuWidth, 0, display_w, display_h);

                // Bind deferred pass 2 shader:
                deferredPass2Shader.bind();

                // Bind position texture from pass 1:
                deferredLightingFBO.getTexture2D(0).bind(0);
                deferredPass2Shader.setUniform("gBufferPosition", 0);

                // Bind normal texture from pass 1:
                deferredLightingFBO.getTexture2D(1).bind(1);
                deferredPass2Shader.setUniform("gBufferNormal", 1);

                // Bind color texture from pass 1:
                deferredLightingFBO.getTexture2D(2).bind(2);
                deferredPass2Shader.setUniform("gBufferMaterial", 2);

                // Bind lights to deferredPass2 shader:
                world.bindLights(view, deferredPass2Shader, lightNum);

                // Render quad mesh:
                quadMesh.render();

                // Blit depth texture from the deferredLightingFBO into the default fbo
                // (we do this to be able to render the light icons in the follwoing correctly,
                // because they do not use deferred lighting):
                glBindFramebuffer(GL_READ_FRAMEBUFFER, deferredLightingFBO.getName());
                glBindFramebuffer(GL_DRAW_FRAMEBUFFER, 0);
                glBlitFramebuffer(0, 0, display_w, display_h, menuWidth, 0, display_w + menuWidth, display_h, GL_DEPTH_BUFFER_BIT, GL_NEAREST);
            }
        }

        // End measuring time:
        glEndQuery(GL_TIME_ELAPSED);

        // Wait until time is available:
        int resultAvailable = 0;
        while (!resultAvailable) {
            glGetQueryObjectiv(timingQuery, GL_QUERY_RESULT_AVAILABLE, &resultAvailable);
        }

        // Rendering time:
        GLuint64 time = 0;
        glGetQueryObjectui64v(timingQuery, GL_QUERY_RESULT, &time);
        worldRenderingTime = (time / 1000000.f) * 0.1f + worldRenderingTime * 0.9f;

        // Delete queries:
        glDeleteQueries(1,&timingQuery);

        // Render light positions with cubes (if they should be shown):
        if(showLights){
            world.renderLightCubes(view, projection, lightNum);
        }

        // Render the GUI and draw it to the screen:
        ImGui::Render();
        ImGui_ImplOpenGL3_RenderDrawData(ImGui::GetDrawData());

        // Swap Buffers:
        glfwSwapBuffers(window);
    }

    // Cleanup
    ImGui_ImplOpenGL3_Shutdown();
    ImGui_ImplGlfw_Shutdown();
    ImGui::DestroyContext();

    glfwDestroyWindow(window);
    glfwTerminate();

    return 0;
}